
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "face.h"
#include "mesh.h"
#include "c3ga_util.h"
#include "surface_point.h"
#include "texture.h"
#include "../shared.src/mt19937-2.h"
#include "../shared.src/constants.h"

face::face() : m_u2(0.0f), m_v2(0.0f), m_w2(0.0f), 
					 m_uv(0.0f), m_vw(0.0f), m_wu(0.0f), 
					 m_u_w2(0.0f) 
{
	for (int i = 0; i < 3; i++)
		m_vtx_idx[i] = 0;

	m_uv_dir[0][0] = m_uv_dir[0][1] = 
		m_uv_dir[1][0] = m_uv_dir[1][1] = 0.0; 
}

face::face(const int vtx_idx[3]) : m_u2(0.0f), m_v2(0.0f), m_w2(0.0f), 
												m_uv(0.0f), m_vw(0.0f), m_wu(0.0f),
												m_u_w2(0.0f) 
{
	for (int i = 0; i < 3; i++)
		m_vtx_idx[i] = vtx_idx[i];

	m_uv_dir[0][0] = m_uv_dir[0][1] = 
		m_uv_dir[1][0] = m_uv_dir[1][1] = 0.0; 
}

face::face(const face &f) : m_u2(f.m_u2), m_v2(f.m_v2), m_w2(f.m_w2), 
										m_uv(f.m_uv), m_vw(f.m_vw), m_wu(f.m_wu), 
										m_u_w2(f.m_u_w2) , m_pl(f.m_pl)
{
	for (int i = 0; i < 3; i++) {
			m_vtx_idx[i] = f.m_vtx_idx[i];
			m_edge_line[i] = f.m_edge_line[i];
		}

	m_uv_dir[0][0] = f.m_uv_dir[0][0];
	m_uv_dir[0][1] = f.m_uv_dir[0][1];
	m_uv_dir[1][0] = f.m_uv_dir[1][0];
	m_uv_dir[1][1] = f.m_uv_dir[1][1];
}


face::~face() {

}


face &face::operator=(const face &f) {
	if (this != &f) {
		for (int i = 0; i < 3; i++) {
			m_vtx_idx[i] = f.m_vtx_idx[i];
			m_edge_line[i] = f.m_edge_line[i];
		}

		m_u2 = f.m_u2, 
		m_v2 = f.m_v2;
		m_w2 = f.m_w2; 
									
		m_uv = f.m_uv; 
		m_vw = f.m_vw; 
		m_wu = f.m_wu; 
									
		m_u_w2 = f.m_u_w2;

		m_uv_dir[0][0] = f.m_uv_dir[0][0];
		m_uv_dir[0][1] = f.m_uv_dir[0][1];
		m_uv_dir[1][0] = f.m_uv_dir[1][0];
		m_uv_dir[1][1] = f.m_uv_dir[1][1];

		m_pl = f.m_pl;
	}
	return *this;
}


void face::precompute_constants(const mesh &m) {
	// compute lines along the edges of the face
	for (int i = 0; i < 3; i++)
		m_edge_line[i] = get_vtx(i, m).get_pt() ^ get_vtx((i + 1) % 3, m).get_pt() ^ c3ga::ni;
		// don't make them unit, this will mess up the interpolation!
		//m_edge_line[i] = unit_r(get_vtx(i, m).get_pt() ^ get_vtx((i + 1) % 3, m).get_pt() ^ c3ga::ni);
	
	// compute the plane of the face
	m_pl = unit_r(get_vtx(0, m).get_pt() ^ get_vtx(1, m).get_pt() ^ get_vtx(2, m).get_pt() ^ c3ga::ni);

	/*
    Compute the direction (dual plane) of each of the 
	three m_edge_line variables are lines along the face.
    */
	c3ga u = ((c3ga::ni ^ c3ga::no) << m_edge_line[0]);
    c3ga v = ((c3ga::ni ^ c3ga::no) << m_edge_line[1]);
    c3ga w = ((c3ga::ni ^ c3ga::no) << m_edge_line[2]);


	// compute interpolation constants
	m_u2 = (u % u).scalar();
	m_v2 = (v % v).scalar();
	m_w2 = (w % w).scalar();

	m_uv = (u % v).scalar();
	m_vw = (v % w).scalar();
	m_wu = (w % u).scalar();

	m_u_w2 = ((u^w) % (u^w)).scalar();
}

bool face::uses_vtx_idx(int vtx_idx) const {
	return ((vtx_idx == get_vtx_idx(0)) || 
		(vtx_idx == get_vtx_idx(1)) || 
		(vtx_idx == get_vtx_idx(2)));
}

// todo: make better use of GA in this problem . . . !!!!
void face::compute_local_texture_basis(const mesh &m) {
	// Compute 'a' and 'b', Euclidean vectors in 
	// texture space along two of the edges of the polygon
	c3ga a = m.get_vertex(m_vtx_idx[1]).get_tex_pt() - m.get_vertex(m_vtx_idx[0]).get_tex_pt();
	c3ga b = m.get_vertex(m_vtx_idx[2]).get_tex_pt() - m.get_vertex(m_vtx_idx[0]).get_tex_pt();

	GAIM_FLOAT a1 = ((c3ga::no << a) % c3ga::e1).scalar();
	GAIM_FLOAT a2 = ((c3ga::no << a) % c3ga::e2).scalar();
	GAIM_FLOAT b1 = ((c3ga::no << b) % c3ga::e1).scalar();
	GAIM_FLOAT b2 = ((c3ga::no << b) % c3ga::e2).scalar();

	GAIM_FLOAT alpha1, beta1, alpha2, beta2;
		
	if ((b2 * a1 - b1 * a2) == 0.0f) { // degenerate triangle; spans no area in texture space
		alpha1 = alpha2 = beta1 = beta2 = 0.0;
	}
	else {
		// invert matrix . . . 
		beta1 = a2 / (b1 * a2 - b2 * a1);
		alpha1 = (a2 == 0.0f) ? 1.0f / a1 : -beta1 * b2 / a2;
		beta2 = a1 / (b2 * a1 - b1 * a2);
		alpha2 = (a1 == 0.0f) ? 1.0f / a2 : -beta2 * b1 / a1;
	}

	m_uv_dir[0][0] = alpha1;
	m_uv_dir[0][1] = beta1;
	m_uv_dir[1][0] = alpha2;
	m_uv_dir[1][1] = beta2;
}

bool face::find_intersection(const c3ga &ray_dual_plane, const c3ga &ray_dual_line, surface_point &spt, bool find_closest) const {
	// check if line intersects triangle:
	GAIM_FLOAT side1 = (c3ga::no % (ray_dual_line << get_edge_line(0))).scalar();
	GAIM_FLOAT side2 = (c3ga::no % (ray_dual_line << get_edge_line(1))).scalar();
	GAIM_FLOAT side3 = (c3ga::no % (ray_dual_line << get_edge_line(2))).scalar();
	int sd;

	// check if face was intersected, and if so, from what side did the ray come:
	if ((side1 > 0.0) && (side2 > 0.0) && (side3 > 0.0)) {
		sd = 1; // front side
	}
	else if ((side1 < 0.0) && (side2 < 0.0) && (side3 < 0.0)) {
		sd = -1; // back side
	}
	else if ((side1 >= 0.0) && (side2 >= 0.0) && (side3 >= 0.0)) {
		sd = 1; // front side (barely touching)
	}
	else return false;

	// compute the intersection point & distance to ray start
	c3ga I = (ray_dual_line << get_pl()) / ((ray_dual_line << get_pl()) % (c3ga::no ^ c3ga::ni));
	GAIM_FLOAT d = -((ray_dual_plane << I) % c3ga::no).scalar();
	if (d < RAY_INTERSECTION_MIN_DISTANCE) return false;

	// if no precise info required, we're done:
	if (!find_closest) {
		spt.set_valid(true);
		spt.set_distance_local(d);
		return true;
	}
	else {
		if ((!spt.get_valid()) ||
			(spt.get_valid() && (d < spt.get_distance_local()))) {
			spt.set_valid(true);
			spt.set_distance_local(d);
			spt.set_pt_local(I);
			spt.set_side(sd);
			return true;
		}
		else return false;
	}
}

void face::compute_interpolation_weights(const c3ga &pt, GAIM_FLOAT iw[3], const vertex &vtx0) const {
	GAIM_FLOAT u_d, v_d, w_d;

    // compute u, v, w as Euclidean vectors:
    c3ga u = (c3ga::no ^ c3ga::ni) << get_edge_line(0);
    c3ga v = (c3ga::no ^ c3ga::ni) << get_edge_line(1);
    c3ga w = (c3ga::no ^ c3ga::ni) << get_edge_line(2);

	// compute local position of point 'pt', as a Euclidean vector
	c3ga d(c3ga:: no<<(pt - (vtx0.get_pt() ^ c3ga::ni)));

    // compute inner product of d with each edge direction
	u_d = (u % d).scalar();
	v_d = (v % d).scalar();
	w_d = (w % d).scalar();

    // compute the weights for the three vertices :
    iw[0] = -(m_v2 * u_d - m_uv * v_d) / m_u_w2 + 1.0f;
    iw[1] = -(m_w2 * v_d - m_vw * w_d) / m_u_w2;
    iw[2] = -(m_u2 * w_d - m_wu * u_d) / m_u_w2;
}

void face::compute_surface_point_details(surface_point &spt, const model &m) const {
	// get the three vertices of this face
	const vertex &vtx0(get_vtx(0, *m.get_mesh()));
	const vertex &vtx1(get_vtx(1, *m.get_mesh()));
	const vertex &vtx2(get_vtx(2, *m.get_mesh()));

	// compute the interpolation weights
	GAIM_FLOAT iw[3]; // iw = interpolation weights
	compute_interpolation_weights(spt.get_pt_local(), iw, vtx0);

	// compute the interpolated texture point
	spt.set_tex_pt(iw[0] * vtx0.get_tex_pt() + iw[1] * vtx1.get_tex_pt() + iw[2] * vtx2.get_tex_pt());

	// compute interpolated, normalized surface normal
	c3ga att(unit_e(iw[0] * vtx0.get_att() + iw[1] * vtx1.get_att() + iw[2] * vtx2.get_att()));

	// apply bumpmap if it is enabled:
	// todo: clean up GA usage in here . . .
	if (m.get_bump_map_texture().enabled()) {
		float bump[3];
		m.get_bump_map_texture().get_bump_map_value(spt.get_tex_pt(), bump);

		// construct attitudes for from the edge lines 0 & 2
		c3ga att_edge_0 = get_edge_line(0).dual() ^ c3ga::ni;
		c3ga att_edge_2 = get_edge_line(2).dual() ^ c3ga::ni;

		// transform bumpmap from the 2D 'bumpmap frame' to the local 3D frame
		c3ga bump_att = m.get_bumpyness() * 
			((m_uv_dir[0][1] * bump[0] + m_uv_dir[1][1] * bump[1]) * att_edge_2 -
			(m_uv_dir[0][0] * bump[0] + m_uv_dir[1][0] * bump[1]) * att_edge_0);

		att = unit_e(att + bump_att);
	}

	// set surface attitude in model frame
	spt.set_att_local(att);
}
