
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <math.h>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <iostream>
#include <string>
#include <sstream>

#include <GL/gl.h>

#include "c3ga_util.h"
#include "mv_lexer.hpp"
#include "mv_parser.hpp"
#include "../shared.src/string_util.h"

namespace c3ga {

mv exp(const mv &x, int order /*= 9*/) {
    int i;
    mv result;

    result = 1.0;
    if (order == 0) {
        return result;
    }

    // scale by power of 2 so that its norm is < 1
    unsigned long max = (unsigned long)x.largestCoordinate();
    unsigned long scale=1;
    if (max > 1) scale <<= 1;
    while (max)
    {
        max >>= 1;
        scale <<= 1;
    }

    mv scaled = x * scalar(1.0f / (mv::Float)scale);

    // taylor approximation
    mv tmp;

    tmp = 1.0;
    for (i = 1; i < order; i++) {
        tmp = tmp*scaled * scalar(1.0f / (mv::Float)i);
        result += tmp;
    }

    // undo scaling
    while (scale > 1)
    {
        result *= result;
        scale >>= 1;
    }
	return result;
}


/// Multiplies the current OpenGL matrix with versor 'v'
void unit_versor_glMult(const EXForm &v) {
	// compute the images of all OpenGL basis vectors
	flatPoint image[4] = {
			_flatPoint(v * e1ni * inverse(v)), // image of e1
			_flatPoint(v * e2ni * inverse(v)), // image of e2
			_flatPoint(v * e3ni * inverse(v)), // image of e3
			_flatPoint(v * noni * inverse(v))  // image of origin
	};
/*
	printf("\nversor: %s\n", v.string());
	printf("versor * no / versor: %s\n", string(v * c3ga::no * inverse(v)));
	printf("image e1: %s\n",  string(image[0]));
	printf("image e2: %s\n",  string(image[1]));
	printf("image e3: %s\n",  string(image[2]));
	printf("image o: %s\n",  string(image[3]));
*/

	// transfer the coordinates to the OpenGL matrix
	GLfloat GLmatrix[16];
	for (int i = 0; i < 4; i++) {
		flatPoint &pt(image[i]);
		GLmatrix[i * 4 + 0] = (GLfloat)_Float(pt & e1ni);
		GLmatrix[i * 4 + 1] = (GLfloat)_Float(pt & e2ni);
		GLmatrix[i * 4 + 2] = (GLfloat)_Float(pt & e3ni);
		GLmatrix[i * 4 + 3] = (GLfloat)_Float(pt & noni);
	}

	/*printf("The matrix has you: \n");
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			printf("%f ", GLmatrix[j * 4 + i]);
		}
		printf("\n");
	}*/

	// multiply current OpenGL matrix:
	glMultMatrixf(GLmatrix);
}

void dissect_point_pair(const pointPair &pp, point &pt1, point &pt2) {
	mv::Float n = (mv::Float)sqrt(_Float(pp % pp));
	dualPlane v1 = _dualPlane(ni << pp);
	dualPlane v2 = _dualPlane(n * v1);
	dualSphere v3 = _dualSphere(v1 << pp);

	pt1 = _point(v3 + v2);
	pt2 = _point(v3 - v2);
}

void dissect_point_pair_normalized_flat(const pointPair &pp, normalizedFlatPoint &fpt1, normalizedFlatPoint &fpt2) {
	mv::Float n = (mv::Float)sqrt(_Float(pp % pp));
	dualPlane v1 = _dualPlane(ni << pp);
	dualPlane v2 = _dualPlane(n * v1);
	dualSphere v3 = _dualSphere(v1 << pp);

	mv::Float scale = 1.0f / (_Float(v2 << ni) + _Float(v3 << ni));
	fpt1 = _normalizedFlatPoint((scale * ni) ^ (v3 + v2));
	fpt2 = _normalizedFlatPoint((scale * ni) ^ (v3 - v2));
}

mv parseMVString(const std::string &_str, int startDL/*=-1*/, int endDL/*=-1*/) {
	std::string str;
	if ((startDL >= 0) && (endDL >= 0))	str = getMVString(_str, (char)startDL, (char)endDL);
	else str = _str;
	try {
		std::istringstream inputStream(str);
		c3ga::mv_lexer lex(inputStream);

		c3ga::mv_parser par(lex);
		return par.multivector();
	} catch (...) {
		throw -1;
	}
}

} // end of namespace c3ga
