
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <math.h>
#include <FL/Fl.h>
#include <set>

#include "c3ga_util.h"
#include "ui.h"
#include "ui_versors.h"
#include "modeling_window.h"

#include <GL/glu.h>

using namespace c3ga;

modeling_window::modeling_window(int x, int y, int w, int h, const char *label) :
	Fl_Gl_Window(x, y, w, h, label),
	m_mouse_mode(MM_NONE),
	m_inside_spaceball_circle(false),
	m_previous_pick_x(0),
	m_previous_pick_y(0) {
	reset_OpenGL_pick();
	reset_restrict_draw_to_object_idx();

	m_sphere_quadric = gluNewQuadric();
	if (m_sphere_quadric == NULL) throw std::string("modeling_window::modeling_window(): gluNewQuadric() failed");
}

	
modeling_window::~modeling_window() {
	if (m_sphere_quadric)
		gluDeleteQuadric(m_sphere_quadric);

}

void modeling_window::setup_GL_projection_matrix() {
	// set viewport:
	glViewport(0, 0, w(), h());

	// get the current camera:
	const camera &C(g_ui_state->get_scene().get_current_camera());

	// reset projection matrix:
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// preprend the picking matrix to the projection matrix, if required:
	if (OpenGL_pick_active()) { 
		GLint viewport[4];
		glGetIntegerv(GL_VIEWPORT, viewport);
		gluPickMatrix(
			m_OpenGL_pick[0], m_OpenGL_pick[1], 
			m_OpenGL_pick[2] * 2 + 1, m_OpenGL_pick[3] * 2 + 1, 
			viewport);
	}

	// compute the field of view height from the current window size and the desired field of view width:
	mv::Float fov_height = (mv::Float)h() / (mv::Float)w() * C.get_fov_width();

	// call glFrustum to setup a projection matrix:
//	mv::Float znear = 0.1, zfar = 100.0;
	glFrustum( glFrustum_near() * -C.get_fov_width() / 2.0,
		 glFrustum_near() * C.get_fov_width() / 2.0,
		 glFrustum_near() * -fov_height / 2.0,
		 glFrustum_near() * fov_height / 2.0,
		 glFrustum_near(), glFrustum_far());

}

void modeling_window::setup_GL_modelview_matrix() {
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/*
	Get the inverse transform of the camera and 
	multiply it with the GL_MODELVIEW matrix  
	*/
	unit_versor_glMult(g_ui_state->get_scene().get_current_camera().get_xfi());
}

void modeling_window::setup_GL_lighting(const scene &S) {
	// enable lighting
	glEnable(GL_LIGHTING);

	// set light model: local viewer
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

	// set global ambient: (TEMP)
	::color(0.0f, 0.0f, 0.0f).glLightModel(GL_LIGHT_MODEL_AMBIENT);

	// get maximum number of lights from OpenGL
	GLint maxLights;
	glGetIntegerv(GL_MAX_LIGHTS, &maxLights);

	// turn on lights (and disable unused lights):
	for (int i = 0; i < maxLights; i++) {
		if (i < S.get_nb_light_sources()) {
			glEnable(GL_LIGHT0 + i);
			S.get_light_source(i).to_OpenGL(GL_LIGHT0 + i);
		}
		else {
			glDisable(GL_LIGHT0 + i);
		}
	}
}
void modeling_window::draw() {
	const locator *selected_object = NULL;

	// get the scene:
	const scene &S(g_ui_state->get_scene());

	// clear draw buffer 
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// enable some stuff we want:
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE); // normalize normals after transform
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	// setup projection & modelview matrices:
	setup_GL_projection_matrix();
	setup_GL_modelview_matrix();

	// setup lighting
	setup_GL_lighting(S);

	// draw all models:
	int name_idx = -1; // 'name_idx' counts the OpenGL names?
	for (int i = 0; i < S.get_nb_models(); i++) {
		name_idx++;

		// check if drawing is restricted:
		if ((m_restrict_draw_to_object_idx >= 0) && 
			(name_idx != m_restrict_draw_to_object_idx)) continue;

		// if picking mode is active, load the 'name' for this model:
		if (OpenGL_pick_active())
			glLoadName(name_idx);

		// draw the model
		S.get_model(i).to_OpenGL();

		// if this model is selected, draw it again, in wireframe
		if (S.get_model(i).get_selected()) {
			selected_object = &(S.get_model(i));

			// enable wireframe:
			enable_wireframe_mode();
			// draw the model:
			S.get_model(i).to_OpenGL();
			// disable wireframe:
			disable_wireframe_mode();
		}
	}

	// display lights, when required by UI
	// todo: todo todo todo
	if (S.display_lights()) {
		for (int i = 0; i < S.get_nb_light_sources(); i++) {
			name_idx++;

			// check if drawing is restricted:
			if ((m_restrict_draw_to_object_idx >= 0) && 
				(name_idx != m_restrict_draw_to_object_idx)) continue;

			// if picking mode is active, load the 'name' for this model:
			if (OpenGL_pick_active())
				glLoadName(name_idx);

			// draw the light
			S.get_light_source(i).to_OpenGL_model();

			// if this model is selected, draw it again, in wireframe
			if (S.get_light_source(i).get_selected()) {
				selected_object = &(S.get_light_source(i));

				// enable wireframe:
				enable_wireframe_mode();
				// draw the light:
				S.get_light_source(i).to_OpenGL_model();
				// disable wireframe:
				disable_wireframe_mode();
			}
		}
	}

	// if mouse mode is object rotate, then draw bouding sphere around selected object
	if ((g_ui_state->get_mouse_mode() == MM_ROTATE) && 
		(selected_object)) {
		// get the bounding sphere:
		dualSphere bs = _dualSphere(-dual(get_selected_object_bounding_sphere()));

		// 'normalize' the sphere
		bs = bs * (1.0f /  -_Float(bs % ni));

		// if picking mode is active, load the 'name' for this bounding sphere:
		if (OpenGL_pick_active()) {
			glLoadName(get_bounding_sphere_GL_name());
		}

		/**
		Multiply current GL matrix with translation to the center of the bounding sphere .
		This does _not_ have to be location of the object transform!
		*/
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glTranslated(_double(bs & e1) / _double(bs & no), 
			_double(bs & e2) / _double(bs & no), 
			_double(bs & e3) / _double(bs & no));

		// enable blending for transparent sphere around object:
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// set material to dull gray
		::color(0.2f, 0.2f, 0.2f, 0.5f).glMaterial(GL_FRONT_AND_BACK, GL_AMBIENT);
		::color(0.8f, 0.8f, 0.8f, 0.5f).glMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
		::color(0.2f, 0.2f, 0.2f, 0.5f).glMaterial(GL_FRONT_AND_BACK, GL_SPECULAR);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (GLfloat)10.0);

		// draw a sphere with required radius
		gluSphere(m_sphere_quadric, 
			sqrt(fabs(_double(bs % bs))),  // radius
			20, 20); // slices, stacks
		
		// disable blending
		glDisable(GL_BLEND);

		// undo transform
		glPopMatrix();
	}

}


int modeling_window::handle(int event) {
	switch (event) {
	case FL_PUSH:
		// register mouse mode
		m_mouse_mode = g_ui_state->get_mouse_mode();
		// inside / outside spaceball circle:
		m_inside_spaceball_circle = inside_spaceball_circle(get_FLTK_mouse_position());

		// rest of the work is done in separate function:
		handle_push();
		return 1;
	case FL_RELEASE:
		handle_release();
		m_mouse_mode = MM_NONE;
		return 1;
	case FL_MOVE:
		handle_move();
		return 1;
	case FL_DRAG:
		// rest of the work is done in separate function:
		handle_drag();
		return 1;
	default:
		return Fl_Gl_Window::handle(event);
	}
}

void modeling_window::handle_push() {
	m_drag_mouse_pos = get_FLTK_mouse_position();

	// if mouse mode is 'select' then call pick() for OpenGL picking
	if (get_mouse_mode() == MM_SELECT) {
		select_object(cycle_pick().first);
	}

	/*
	if mouse mode is 'object manipulation', then select if the mouse if over some object, 
	otherwise keep current selection
	*/
	if ((get_mouse_mode() == MM_TRANSLATE_ORTHOGONAL) ||
		(get_mouse_mode() == MM_TRANSLATE_PARALLEL) ||
		(get_mouse_mode() == MM_SCALE)) {
		int idx = cycle_pick().first;
		if (idx >= 0) select_object(idx);

		/*
		If mouse mode is translation orthogonal or parallel, register z depth
		*/
		if ((get_mouse_mode() == MM_TRANSLATE_ORTHOGONAL) ||
			(get_mouse_mode() == MM_TRANSLATE_PARALLEL)) {
			m_selection_depth = get_world_z(idx, Fl::event_x(), h() - Fl::event_y());
		}
	}

	// if mouse mode is orbit, remember the versor (translation part only!) we orbit:
	if (get_mouse_mode() == MM_CAMERA_ORBIT) {
		try {
			// get current locator (selected object)
			locator &L(g_ui_state->get_scene().get_selected_object());

			// set orbit  versor:
			m_orbit_versor = L.get_xf();//L.get_translator();
		}
		catch (std::string) { // this exception is thrown when no locator is currently selected
			// get the current camera
			camera &C(g_ui_state->get_scene().get_current_camera());

			// set orbit  versor to camera - 3 e3
			m_orbit_versor =  _EXForm(C.get_xf() * exp(3.0 * e3 ^ ni));
		}
	}

	if (get_mouse_mode() == MM_ROTATE) {
		/*
		Check if the mouse is over the bounding sphere that is used
		to indicate translation.
		If not, select the object the mouse _was_ over
		*/
		int idx = cycle_pick().first;
		if ((idx != get_bounding_sphere_GL_name()) &&
			(idx >= 0)) select_object(idx);

		// remember bounding sphere of selected object:
		try {
			// update the 'rotate point'
			m_previous_rotate_point = get_rotate_point(); // takes all require info from 'this' and FLTK state
		} catch(std::string &) {/*Thrown when no object is selected*/}

	}


}


void modeling_window::handle_release() {
}

void modeling_window::handle_move() {
}

void modeling_window::handle_drag() {
	vectorE2GA mouse_motion = _vectorE2GA(get_FLTK_mouse_position() - m_drag_mouse_pos);

	// get the current camera
	camera &C(g_ui_state->get_scene().get_current_camera());

	switch (get_mouse_mode()) {
	case MM_CAMERA_TRANSLATE_PARALLEL:
		C.post_mul_xf(create_camera_translation_versor_parallel(_vectorE2GA(0.01f * mouse_motion)));
		g_ui_state->scene_changed();
		break;
	case MM_CAMERA_TRANSLATE_ORTHOGONAL:
		C.post_mul_xf(create_camera_translation_versor_orthogonal(_vectorE2GA(0.01f * mouse_motion)));
		g_ui_state->scene_changed();
		break;
	case MM_CAMERA_ROTATE:
		{
			// get the transform we want to 'add' to the camera transform
			EXForm xf;
			if (m_inside_spaceball_circle)
				xf = create_camera_rotation_versor_outside_screen_plane(_vectorE2GA(0.001f * mouse_motion));
			else xf = create_camera_rotation_versor_in_screen_plane(
				_vectorE2GA((get_FLTK_mouse_position() - get_window_center()) * (1.0f / (-0.3f * w() * h()))), mouse_motion);

			// simply orbit the current camera:
			C.post_mul_xf(xf);

			g_ui_state->scene_changed();
			break;
		}
	case MM_CAMERA_ORBIT:
		{
			// get the translator (in camera frame) from camera to orbit versor 
			freeVector t = _freeVector(C.get_xfi() * get_orbit_versor() * noni * inverse(get_orbit_versor()) * C.get_xf() - 
				noni);
			EXForm T = _EXForm(exp(-scalar(0.5) * t));

			// get the transform we want to 'add' to the camera transform
			EXForm xf;
			if (m_inside_spaceball_circle)
				xf = create_camera_rotation_versor_outside_screen_plane(
				_vectorE2GA(0.005f * mouse_motion));
			else xf = create_camera_rotation_versor_in_screen_plane(
				_vectorE2GA((get_FLTK_mouse_position() - get_window_center()) * (1.0f / (-0.3f * w() * h()))), mouse_motion);

			/*
			post multiply the camera transform with the rotation 'xf' around 
			the orbit versor
			*/
			C.post_mul_xf(_EXForm(T * xf * inverse(T)));

			g_ui_state->scene_changed();
			break;
		}
	case MM_CAMERA_FOV:
		C.set_fov_width(C.get_fov_width() - 0.01f * _Float(mouse_motion << e2));
		g_ui_state->scene_changed();
		break;
	case MM_TRANSLATE_ORTHOGONAL:
	case MM_TRANSLATE_PARALLEL:
	case MM_SCALE:
	case MM_ROTATE:
		{
			try {
				// get current locator (selected object)
				locator &L(g_ui_state->get_scene().get_selected_object());

				// translate orthogonal
				if (get_mouse_mode() == MM_TRANSLATE_ORTHOGONAL) {
					L.pre_mul_xf(create_translation_versor_orthogonal(C.get_xf(), L.get_xf(), mouse_motion));
				}
				// translate parallel
				else if (get_mouse_mode() == MM_TRANSLATE_PARALLEL) {
					/*
					The second argument for create_translation_versor_parallel() is the mouse_motion,
					but scaled for the distance to the point where the mouse 'grabbed' the object.
					*/
					vectorE2GA scaled_mouse_motion_in_camera_frame = _vectorE2GA(mouse_motion * (get_selection_depth() * C.get_fov_width() / w()));
					L.pre_mul_xf(create_translation_versor_parallel(C.get_xf(), scaled_mouse_motion_in_camera_frame));
				}
				// scale 
				else if (get_mouse_mode() == MM_SCALE) {
					if (L.can_scale())
//						L.pre_mul_xf(create_scaling_versor(L.get_xf(), mouse_motion));
						L.post_mul_xf(create_scaling_versor(mouse_motion));
				}
				// rotate 
				else if (get_mouse_mode() == MM_ROTATE) {
					// get new point on bounding sphere / model plane
					dualSphere pt = get_rotate_point(); // takes all required info from 'this' and FLTK state
//					printf("Rotate point: %s\n", c3ga::string(pt));
					// get bounding sphere
					dualSphere s = _dualSphere(dual(get_selected_object_bounding_sphere()));

					// are the points on the sphere?
					bool on_bs1 = fabs(_Float(pt % s)) < 10e-6f;
					bool on_bs2 = fabs(_Float(get_previous_rotate_point() % s)) < 10e-6f;

					if (_Float(pt << ni) &&  // check if 'pt' is valid
						_Float(get_previous_rotate_point() << ni) && // check if prev. 'pt' is valid
						((on_bs1 ^ on_bs2) == 0)) { // check if both on sphere, or both on plane

						line L1 = _line(s ^ pt ^ ni);
						line L2 = _line(s ^ (pt + get_previous_rotate_point()) ^ ni);

						// TODO: this rules, but is it also stable?:
						L.pre_mul_xf(_EXForm(L1 * L2 * (1.0 / sqrt(fabs(_double(norm_e2(L1 * L2)))))));
					}

					// update the 'rotate point'
					m_previous_rotate_point = pt; 
				}
				g_ui_state->scene_changed();

			}
			catch (std::string S) {} // an exception is thrown if no object is currently selected
		}
		break;
	case 	MM_SPOTLIGHT:
		try {
			light_source &L(g_ui_state->get_scene().get_selected_light_source());
			L.set_spot_cutoff(L.get_spot_cutoff() + 0.2f * _Float(mouse_motion % e2));
			g_ui_state->scene_changed();
		}
		catch (std::string S) {} // an exception is thrown if no light source is currently selected
		break;
	}

	m_drag_mouse_pos = get_FLTK_mouse_position();
}

mouse_modes modeling_window::get_mouse_mode() const {
	return m_mouse_mode;
}	

vectorE2GA modeling_window::get_FLTK_mouse_position() const {
	return _vectorE2GA(Fl::event_x() * e1 + (h() - Fl::event_y()) * e2);
}

vectorE2GA modeling_window::get_window_center() const {
	/*
	The calls to Fl::event_x() and Fl::event_y() return the coordinates 
	of the mouse in pixels. We reflected the y-coordinate by subtracting
	it from the height (returned h()) of the window.
	*/
	return _vectorE2GA(0.5 * w() * e1 + 0.5 * h() * e2);
}

bool modeling_window::inside_spaceball_circle(const vectorE2GA &pos) {
	const mv::Float inner_spaceball_size = 0.375;
	int s = w();
	if (h() < s) s = h();

	return (-2.0f * _Float(cga_point(pos) << cga_point(get_window_center())) < 
		(inner_spaceball_size * s * inner_spaceball_size * s));
}


void modeling_window::reset_OpenGL_pick() {
	m_OpenGL_pick[0] = m_OpenGL_pick[1] = m_OpenGL_pick[2] = m_OpenGL_pick[3] = -1;
}

bool  modeling_window::OpenGL_pick_active() const {
	return ((m_OpenGL_pick[0] >= 0) && (m_OpenGL_pick[1] >= 0));
}

std::vector<std::pair<GLuint, GLuint> >  modeling_window::pick(int x, int y, int w, int h) {
	// activate OpenGL context (since this function is from the event handler)
	make_current();

	// get scene
	const scene &S(g_ui_state->get_scene());

	// set picking data
	m_OpenGL_pick[0] = x;
	m_OpenGL_pick[1] = y;
	m_OpenGL_pick[2] = w;
	m_OpenGL_pick[3] = h;
	
	// count the number of objects
	int nb_of_objects = S.get_nb_light_sources() + S.get_nb_models() + 1;

	// allocate select buffer
	std::vector<GLuint> sb;
	sb.resize(nb_of_objects * 4);

	// set select buffer, set render mode to select
	glSelectBuffer(nb_of_objects * 4 * sizeof(GLuint), &(sb[0]));
	glRenderMode(GL_SELECT);

	// initialize OpenGL picking names 
	glInitNames();
	glPushName(0);

	// call the normal drawing function (it will call glLoadName() to load the right names for models / lights)
	draw();

	// get the number of hit records by returning to normal 'render mode'
	int nb_hits = glRenderMode(GL_RENDER);

	// reset m_OpenGL_pick
	reset_OpenGL_pick();


	/*
	collect the hits
	OpenGL hits records are:
	-number of names on the name stack (we make sure that we have only one name on the stack)
	-minimum z of hit
	-maximum z of hit
	-name of object 

	Z-sort them by inserting them into a set.
	*/
	std::set<std::pair<GLuint, GLuint> > _hits;
	for (int i = 0; i < nb_hits; i++) {
		int idx =  i * 4;

		// check hit record (we should have only one name on the stack
		if (sb[idx] != 1) {break;} // error!

		// register hit:
		_hits.insert(std::make_pair<GLuint, GLuint>((sb[idx + 1]/2) + (sb[idx + 2]/2), sb[idx + 3]));
	}

	// insert hits into vector (& swap order of first and second)
	std::vector<std::pair<GLuint, GLuint> > hits;
	for (std::set<std::pair<GLuint, GLuint> >::const_iterator I = _hits.begin();
		I != _hits.end(); I++)
		hits.push_back(std::make_pair<GLuint, GLuint>(I->second, I->first));

	// done
	return hits;
}

std::pair<int, int> modeling_window::cycle_pick() {
	std::vector<std::pair<GLuint, GLuint> > hits = pick(Fl::event_x(), h() - Fl::event_y(), 1, 1);

	unsigned int pick_idx = 0; // the final 'pick' that we select
	if (hits.size() > 1) { 
		// only keep the previous pick hits if x and y of event are unchanged:
		if ((Fl::event_x() != m_previous_pick_x) ||
			(Fl::event_y() != m_previous_pick_y))
			m_previous_pick_hits.clear();

		// implement 'cycling' so we can pick objects behind other objects:
		while (pick_idx < hits.size()) {
			if ((pick_idx < m_previous_pick_hits.size()) &&
				(hits[pick_idx].first == m_previous_pick_hits[pick_idx]))
				pick_idx++;
			else break;
		}
		if (pick_idx == hits.size()) pick_idx = 0;

		// remember this hit for next time:
		m_previous_pick_hits.clear();
		for (unsigned int i = 0; i <= pick_idx; i++) {
			m_previous_pick_hits.push_back(hits[i].first);
		}
	}
	else m_previous_pick_hits.clear();
	m_previous_pick_x = Fl::event_x();
	m_previous_pick_y = Fl::event_y();

	// return hit 
	return (pick_idx < hits.size()) 
		? std::make_pair<int, int>(hits[pick_idx].first, hits[pick_idx].second) 
		: std::make_pair<int, int>(-1, -1);
}

void modeling_window::select_object(int idx) {
	g_ui_state->scene_changed();
	if (idx < 0) {
		g_ui_state->get_scene().reset_selected_object();
		return;
	}
	else {
		g_ui_state->get_scene().set_selected_object(idx);
	}

}

/// Converts the z as returned by a picking record to a real z coordinate
GLfloat modeling_window::get_world_z(int object_idx, int x, int y) {
	// make sure picking is turned off
	reset_OpenGL_pick();

	// render & read from back buffer (so nothing is visible to the user)
	glDrawBuffer(GL_BACK);
	glReadBuffer(GL_BACK);

	// write to depth buffer, not to color buffer
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask(GL_TRUE);

	// restrict drawing to single object:
	m_restrict_draw_to_object_idx = object_idx;

	// draw the scene:
	draw();

	// set color mask back to normal write 
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	// reset drawing restriction
	reset_restrict_draw_to_object_idx();

	// get pixel depth at that location
	GLfloat z_pixel_window_coordinates, z_pixel_world_coordinates = 1.0f;
	glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &z_pixel_window_coordinates);

	// do this conversion:
	if (z_pixel_window_coordinates < 1.0) { // zPixelWindowCoordinates == 1.0 this means that you probably missed the object
		// undo OpenGL transformations:
		z_pixel_world_coordinates = (GLfloat)(-glFrustum_far()*glFrustum_near()/(glFrustum_far()-glFrustum_near()) / 
			(z_pixel_window_coordinates - 0.5f * (glFrustum_far()+glFrustum_near())/(glFrustum_far()-glFrustum_near()) - 0.5f));
	}
	return z_pixel_world_coordinates;
}

void modeling_window::reset_restrict_draw_to_object_idx() {
	m_restrict_draw_to_object_idx = -1;
}

GLfloat modeling_window::get_selection_depth() const {
	return m_selection_depth;
}

const EXForm &modeling_window::get_orbit_versor() const {
	return m_orbit_versor;
}

int modeling_window::get_bounding_sphere_GL_name() const {
	const scene &S(g_ui_state->get_scene());
	return S.get_nb_models() + S.get_nb_light_sources();
}

sphere modeling_window::get_selected_object_bounding_sphere() const {
	// remember bounding sphere of selected object:
	try {
		const locator &L = g_ui_state->get_scene().get_selected_object();
		try {
			// if model is selected: get it's bounding sphere:
			return /*(L.get_xf() * 
				(g_ui_state->get_scene().get_selected_model().get_mesh()->get_bounding_sphere()).dual() * 
				L.get_xf().inverse())(GRADE4);*/
				g_ui_state->get_scene().get_selected_model().get_bounding_sphere();
		}
		catch (std::string &) { //
			// otherwise a light is selected, just make up a bounding sphere
			return _sphere(L.get_xf() * dual((no - 0.1f * ni)) * L.get_xfi());
		}
	} catch(std::string &) {/*Thrown when no object is selected (should never happen*/
		return sphere();
	}

}


dualSphere modeling_window::compute_intersection_location(const camera &C, 
												 const vectorE2GA &mouse_position,
												 const sphere &bounding_sphere) const {	
	// compute the center of the camera in the global frame
	point camera_center = _point(C.apply_xf_to(no));

	/*
	Compute the line that goes from the camera center through the pixel
	that the mouse was over,  in the global frame:
	*/
	line l = _line(C.apply_xf_to(no ^ (0.5 * C.get_fov_width()  * mouse_position - e3) ^ ni));

	// compute intersection
	pointPair I = _pointPair(dual(l) << bounding_sphere);

	// test for real intersection
	if (_Float(I << I) < 0) {
		// compute plane parallel to viewport, at position of bounding sphere:
		plane p = _plane(dual(bounding_sphere) ^ 
			C.apply_xf_to(e1 ^ e2 ^ ni));

		// return the sphere at the intersection of the line and the plane:
		return _dualSphere(-(no << (dual(l) << p)));
	}

	// dissect the point pair
	point p1, p2;
	dissect_point_pair(I, p1, p2);

	// return the closest point
	if (-_Float(camera_center << p1) < -_Float(camera_center << p2))
		return _dualSphere(p1); // todo: this type of conversion is _not_ optimized . . . 
	else return _dualSphere(p2); // todo: this type of conversion is _not_ optimized . . . 
}


dualSphere modeling_window::get_rotate_point() const {
	// get the camera
	const camera &C(g_ui_state->get_scene().get_current_camera());

	/*
	Compute the point where the ray from the camera through the mouse
	intersect the bounding sphere:
	*/
	try {
		// set the rotate point
		return compute_intersection_location(C, 
			_vectorE2GA(
			(get_FLTK_mouse_position() - get_window_center()) * (1.0f / (0.5f * w()))), get_selected_object_bounding_sphere());
	}
	catch (std::string) {
		// return invalid the rotate point:
		return dualSphere();
	}
}

const dualSphere &modeling_window::get_previous_rotate_point() const {
	return m_previous_rotate_point;
}

void modeling_window::enable_wireframe_mode() const {
	glDisable(GL_LIGHTING);
	glColor3f(1.0, 1.0, 1.0);
	glPolygonMode(GL_FRONT, GL_LINE);
	glEnable(GL_POLYGON_OFFSET_LINE);
	glPolygonOffset(-0.5, -0.5);
}

void modeling_window::disable_wireframe_mode() const {
	glDisable(GL_POLYGON_OFFSET_LINE);
	glPolygonMode(GL_FRONT, GL_FILL);
	glEnable(GL_LIGHTING);
}

