
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _E3LA_H_
#define _E3LA_H_

#include <stdio.h>
#include <math.h>

#ifndef GAIM_FLOAT
#define GAIM_FLOAT float
#endif 


#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

class e3la_vec;
class e3la_mat;
class e3la_quat;
class e3la_axis;
class e3la_euler;

class e3la_vec {
public:
	e3la_vec() : x(0.0), y(0.0), z(0.0) {}; // null constructor
	e3la_vec(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) : x(ax), y(ay), z(az) {}; // parametarised constructor
	e3la_vec(const e3la_vec &v) : x(v.x), y(v.y), z(v.z) {}; // copy constructor
	~e3la_vec() {};

	e3la_vec& operator=(const e3la_vec &v) { // assignment operator
		x = v.x; y = v.y; z = v.z;
		return *this;
    }

	// setting values
	inline void null() {x = y = z = 0.0;}
	inline void set(const GAIM_FLOAT *c) {x = c[0]; y = c[1]; z = c[2];}
	inline void set(GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) {x = ax; y = ay; z = az;}

	e3la_vec operator+(const e3la_vec &v) const; // addition operator
	e3la_vec operator-(const e3la_vec &v) const; // subtraction operator
	e3la_vec operator-() const; // negation operator
	e3la_vec operator+() const; // nothing operator
	GAIM_FLOAT operator%(const e3la_vec &v) const; // dot product operator
	e3la_vec operator^(const e3la_vec &v) const; // outer product operator
	inline e3la_vec operator/(GAIM_FLOAT d) const { // division operator
		e3la_vec r(x / d, y / d, z / d); return r; }; 

	e3la_vec &operator+=(const e3la_vec &v); // addition operator
	e3la_vec &operator-=(const e3la_vec &v); // addition operator

	inline GAIM_FLOAT norm() {return (GAIM_FLOAT)sqrt(x * x + y * y + z * z);}
	inline GAIM_FLOAT norm2() {return x * x + y * y + z * z;}
	e3la_vec& normalize(); // normalize this

	void print(const char *text, const char *format = NULL) const; // print text in 'format' (format default %2.2f)

	GAIM_FLOAT x, y, z; // the coordinates
};

inline e3la_vec operator*(GAIM_FLOAT s, const e3la_vec &v) {
	e3la_vec r(v.x *s, v.y *s, v.z *s);
	return r;
}

inline e3la_vec operator*(const e3la_vec &v, GAIM_FLOAT s) {
	return operator*(s, v);
}

class e3la_quat {
public:
	e3la_quat() : s(1.0), x(0.0), y(0.0), z(0.0) {}; // identity constructor
	e3la_quat(GAIM_FLOAT as, GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) : s(as), x(ax), y(ay), z(az) {}; // parametarised constructor
	e3la_quat(const e3la_quat &q) : s(q.s), x(q.x), y(q.y), z(q.z) {}; // copy constructor
	~e3la_quat() {};

	inline e3la_quat& operator=(const e3la_quat &q) { // assignment operator
		s = q.s; x = q.x; y = q.y; z = q.z;
		return *this;
    }

	void set(char axis, GAIM_FLOAT angle); // set to rotation about 'x', 'y' or 'z' axis
	void set(const e3la_mat &m); // auto conversion from matrix
	void set(const e3la_axis &a); // auto conversion from axis
	void set(const e3la_euler &ea); // auto conversion from euler angle sequence
	void identity(); // set to identity

	e3la_quat& operator=(const e3la_mat &m) {set(m); return *this;}; // assignment operator (auto conversion from matrix)
	e3la_quat& operator=(const e3la_axis &a) {set(a); return *this;}; // assignment operator (auto conversion from axis)
	e3la_quat& operator=(const e3la_euler &ea) {set(ea); return *this;}; // assignment operator (auto conversion from euler angle sequence)

	e3la_quat operator*(const e3la_quat &q) const; // quaternion multiplication operator
	e3la_vec operator*(const e3la_vec &v) const; // vector multiplication operator

	e3la_quat operator!() const; // inversion operator

	e3la_quat& normalize(); // normalize this

	void print(const char *text, const char *format = NULL) const; // print text in 'format' (format default %2.2f)

	GAIM_FLOAT s, x, y, z; // the coordinates
};

class e3la_mat {
public:
	e3la_mat() : m11(1.0), m12(0.0), m13(0.0), m21(0.0), m22(1.0), m23(0.0), m31(0.0), m32(0.0), m33(1.0) {}; // identity constructor
	e3la_mat(const e3la_mat &m) : m11(m.m11), m12(m.m12), m13(m.m13), m21(m.m21), m22(m.m22), m23(m.m23), m31(m.m31), m32(m.m32), m33(m.m33) {}; // copy constructor
	e3la_mat(GAIM_FLOAT am11, GAIM_FLOAT am12, GAIM_FLOAT am13,
				GAIM_FLOAT am21, GAIM_FLOAT am22, GAIM_FLOAT am23,
				GAIM_FLOAT am31, GAIM_FLOAT am32, GAIM_FLOAT am33) : m11(am11), m12(am12), m13(am13), m21(am21), m22(am22), m23(am23), m31(am31), m32(am32), m33(am33) {}; // copy constructor
	~e3la_mat() {};

	inline e3la_mat& operator=(const e3la_mat &m) { // assignment operator
		m11 = m.m11; m12 = m.m12; m13 = m.m13; 
		m21 = m.m21; m22 = m.m22; m23 = m.m23; 
		m31 = m.m31; m32 = m.m32; m33 = m.m33;
		return *this;
    }

	void set(char axis, GAIM_FLOAT angle); // set to rotation about 'x', 'y' or 'z' axis
	void set(const e3la_quat &q); // auto conversion from quaternion
	void set(const e3la_axis &a); // auto conversion from axis
	void set(const e3la_euler &ea); // auto conversion from euler angle sequence
	void identity(); // set to identity

	e3la_mat& operator=(const e3la_quat &q) {set(q); return *this;}; // assignment operator (auto conversion from quaternion)
	e3la_mat& operator=(const e3la_axis &a) {set(a); return *this;}; // assignment operator (auto conversion from axis)
	e3la_mat& operator=(const e3la_euler &ea) {set(ea); return *this;}; // assignment operator (auto conversion from euler angle sequence)

	e3la_mat operator*(const e3la_mat &m) const; // matrix multiplication operator
	e3la_vec operator*(const e3la_vec &v) const; // vector multiplication operator

	e3la_mat operator!() const; // inversion operator

	void print(const char *text, const char *format = NULL) const; // print text in 'format' (format default %2.2f)

	// the coordinates
	GAIM_FLOAT m11, m12, m13;
	GAIM_FLOAT m21, m22, m23;
	GAIM_FLOAT m31, m32, m33;
};

class e3la_axis {
public:
	e3la_axis() : a(0.0), x(1.0), y(0.0), z(0.0) {}; // identity constructor
	e3la_axis(GAIM_FLOAT aa, GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) : a(aa), x(ax), y(ay), z(az) {}; // parametarised constructor
	e3la_axis(const e3la_axis &a) : a(a.a), x(a.x), y(a.y), z(a.z) {}; // copy constructor
	~e3la_axis() {};

	void set(const e3la_mat &m); // auto conversion from matrix
	void set(const e3la_quat &q); // auto conversion from quaternion
	inline void set(GAIM_FLOAT aa, GAIM_FLOAT ax, GAIM_FLOAT ay, GAIM_FLOAT az) {a = aa; x = ax; y = ay; z = az;}
	inline void set(GAIM_FLOAT aa, e3la_vec &v) {a = aa; x = v.x; y = v.y; z = v.z;}
	void identity(); // set to identity

	e3la_axis& operator=(const e3la_quat &q) {set(q); return *this;}; // assignment operator (auto conversion from quaternion)
	e3la_axis& operator=(const e3la_mat &m) {set(m); return *this;}; // assignment operator (auto conversion from matrix)

	e3la_vec operator*(const e3la_vec &v) const; // vector multiplication operator

	void print(const char *text, const char *format = NULL) const; // print text in 'format' (format default %2.2f)

	// the coordinates
	GAIM_FLOAT a, x, y, z;
};


typedef enum {
	EULER_XYZ =  1,
	EULER_ZYX,
	EULER_YZX,
	EULER_XZY,
	EULER_ZXY,
	EULER_YXZ
} eulerSeqOrder; 

class e3la_euler {
public:
	e3la_euler() : o(EULER_XYZ), ea1(0.0), ea2(0.0), ea3(0.0) {}; // identity constructor
	e3la_euler(eulerSeqOrder ao, GAIM_FLOAT aea1, GAIM_FLOAT aea2, GAIM_FLOAT aea3) : o(ao), ea1(aea1), ea2(aea2), ea3(aea3) {}; // parametarised constructor
	e3la_euler(const e3la_euler &ea) : o(ea.o), ea1(ea.ea1), ea2(ea.ea2), ea3(ea.ea3) {}; // copy constructor
	~e3la_euler() {};

	void set(eulerSeqOrder order, const e3la_mat &m, const e3la_euler *pea = NULL); // auto conversion from matrix (with continuity, if 'pea' (previous euler angle) provided)
	void set(eulerSeqOrder order, const e3la_quat &q, const e3la_euler *pea = NULL); // auto conversion from quaternion (with continuity, if 'pea' (previous euler angle) provided)
	void identity(); // set to identity

	e3la_euler& operator=(const e3la_quat &q) {set(o, q); return *this;}; // assignment operator (auto conversion from quaternion)
	e3la_euler& operator=(const e3la_mat &m) {set(o, m); return *this;}; // assignment operator (auto conversion from matrix)

	e3la_vec operator*(const e3la_vec &v) const; // vector multiplication operator

	//todo
	//void print(const char *text, const char *format = NULL); // print text in 'format' (format default %2.2f)

	// the coordinates
	eulerSeqOrder o;
	GAIM_FLOAT ea1, ea2, ea3;
};


// utility functions
int intersectLineSphere(const e3la_vec &pt, const e3la_vec &dir, const e3la_vec &spherePt, GAIM_FLOAT sphereSize);
int intersectLineSpherePoints(const e3la_vec &pt, const e3la_vec &dir, const e3la_vec &spherePt, GAIM_FLOAT sphereSize, e3la_vec isPt[2]);

inline e3la_vec e3la_vec::operator+(const e3la_vec &v) const {
	return e3la_vec(x + v.x, y + v.y, z + v.z);
}

inline e3la_vec e3la_vec::operator-(const e3la_vec &v) const {
	return e3la_vec(x - v.x, y - v.y, z - v.z);
}

inline GAIM_FLOAT e3la_vec::operator%(const e3la_vec &v) const {
	return x * v.x + y * v.y + z * v.z;
}

inline e3la_vec e3la_vec::operator^(const e3la_vec &v) const {
	return e3la_vec(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x);
}

inline e3la_vec& e3la_vec::normalize() {
	GAIM_FLOAT l = (GAIM_FLOAT)sqrt(x * x + y * y + z * z);
	x /= l; y /= l; z /= l; 
	return *this;
}

inline e3la_vec &e3la_vec::operator+=(const e3la_vec &v) {
	x += v.x; y += v.y; z += v.z;
	return *this;
}
inline e3la_vec &e3la_vec::operator-=(const e3la_vec &v) {
	x -= v.x; y -= v.y; z -= v.z;
	return *this;
}

inline e3la_vec e3la_vec::operator-() const {
	return e3la_vec(-x, -y, -z);
}

inline e3la_vec e3la_vec::operator+() const {
	return e3la_vec(x, y, z);
}

inline void e3la_vec::print(const char *text, const char *fmt /* = NULL */) const{
	char buf[1024];
	if (text) printf("%s", text);
	if (fmt == NULL) fmt = "%2.2f";
	sprintf(buf, "x: %s y: %s z: %s\n", fmt, fmt, fmt);
	printf(buf, x, y, z);
}

inline void e3la_quat::set(char axis, GAIM_FLOAT angle) {
	int A = (axis == 'x') ? 0 : (axis == 'y') ? 1 : 2;
	s = (GAIM_FLOAT)cos(angle * 0.5);
	GAIM_FLOAT sa = (GAIM_FLOAT)sin(angle * 0.5);
	x = (GAIM_FLOAT)(A == 0) * sa;
	y = (GAIM_FLOAT)(A == 1) * sa;
	z = (GAIM_FLOAT)(A == 2) * sa;	
}

inline void e3la_quat::set(const e3la_mat &m) {
	GAIM_FLOAT qs2, qx2, qy2, qz2;  /* squared magniudes of quaternion components */
	GAIM_FLOAT tmp;
	int i;

	/* 
	first compute squared magnitudes of quaternion components - at least one
	will be greater than 0 since quaternion is unit magnitude
	*/
	qs2 = (GAIM_FLOAT)0.25 * (m.m11 + m.m22 + m.m33 + (GAIM_FLOAT)1.0);
	qx2 = qs2 - (GAIM_FLOAT)0.5 * (m.m22 + m.m33);
	qy2 = qs2 - (GAIM_FLOAT)0.5 * (m.m33 + m.m11);
	qz2 = qs2 - (GAIM_FLOAT)0.5 * (m.m11 + m.m22);

	/* find maximum magnitude component */
	i = (qs2 > qx2 ) ?
	((qs2 > qy2) ? ((qs2 > qz2) ? 0 : 3) : ((qy2 > qz2) ? 2 : 3)) :
	((qx2 > qy2) ? ((qx2 > qz2) ? 1 : 3) : ((qy2 > qz2) ? 2 : 3));

	/* compute signed quaternion components using numerically stable method */
	switch(i) {
	case 0:
		s = (GAIM_FLOAT)sqrt(qs2);
		tmp = (GAIM_FLOAT)0.25 / s;
		x = (m.m32 - m.m23) * tmp;
		y = (m.m13 - m.m31) * tmp;
		z = (m.m21 - m.m12) * tmp;
		break;
	case 1:
		x = (GAIM_FLOAT)sqrt(qx2);
		tmp = (GAIM_FLOAT)0.25 / x;
		s = (m.m32 - m.m23) * tmp;
		y = (m.m12 + m.m21) * tmp;
		z = (m.m13 + m.m31) * tmp;
		break;
	case 2:
		y = (GAIM_FLOAT)sqrt(qy2);
		tmp = (GAIM_FLOAT)0.25 / y;
		s = (m.m13 - m.m31) * tmp;
		z = (m.m23 + m.m32) * tmp;
		x = (m.m21 + m.m12) * tmp;
		break;
	case 3:
		z = (GAIM_FLOAT)sqrt(qz2);
		tmp = (GAIM_FLOAT)0.25 / z;
		s = (m.m21 - m.m12) * tmp;
		x = (m.m31 + m.m13) * tmp;
		y = (m.m32 + m.m23) * tmp;
		break;
	}
	/* for consistency, force positive scalar component [ (s; v) = (-s; -v) ] */
	if (s < 0) {
		s = -s;
		x = -x;
		y = -y;
		z = -z;
	}
	/* normalize, just to be safe */
	tmp = (GAIM_FLOAT)1.0 / (GAIM_FLOAT)sqrt(s*s + x*x + y*y + z*z);
	s *= tmp;
	x *= tmp;
	y *= tmp;
	z *= tmp;
}

inline void e3la_quat::set(const e3la_axis &a) {
	s = (GAIM_FLOAT)cos(a.a * 0.5);
	GAIM_FLOAT sa = (GAIM_FLOAT)sin(a.a * 0.5);
	x = a.x * sa;
	y = a.y * sa;
	z = a.z * sa;
}

inline void e3la_quat::set(const e3la_euler &ea)  {
	e3la_quat q1, q2, q3, q4;

	switch(ea.o) {
	case EULER_XZY:
		q1.set('x', ea.ea1);
		q2.set('z', ea.ea2);
		q3.set('y', ea.ea3);
		break;
	case EULER_YXZ:
		q1.set('y', ea.ea1);
		q2.set('x', ea.ea2);
		q3.set('z', ea.ea3);
		break;
	case EULER_ZXY:
		q1.set('z', ea.ea1);
		q2.set('x', ea.ea2);
		q3.set('y', ea.ea3);
		break;
	case EULER_ZYX:
		q1.set('z', ea.ea1);
		q2.set('y', ea.ea2);
		q3.set('x', ea.ea3);
		break;
	case EULER_XYZ:
		q1.set('x', ea.ea1);
		q2.set('y', ea.ea2);
		q3.set('z', ea.ea3);
		break;
	case EULER_YZX:
		q1.set('y', ea.ea1);
		q2.set('z', ea.ea2);
		q3.set('x', ea.ea3);
		break;
	}

	*this = q3 * q2 * q1;
}

inline void e3la_quat::identity() {
	s = 1.0;
	x = y = z = 0.0;
}

inline e3la_quat e3la_quat::operator*(const e3la_quat &q) const {
	e3la_quat r(s * q.s - x * q.x - y * q.y - z * q.z,
		s * q.x + x * q.s + y * q.z - z * q.y,
		s * q.y - x * q.z + y * q.s + z * q.x,
		s * q.z + x * q.y - y * q.x + z * q.s);
	return r;
}

inline e3la_vec e3la_quat::operator*(const e3la_vec &v) const {
	GAIM_FLOAT x1[3], x2[3], qs;

	x1[0] = y * v.z - z * v.y; 
	x1[1] = z * v.x - x * v.z; 
	x1[2] = x * v.y - y * v.x;

	x2[0] = y * x1[2] - z * x1[1]; 
	x2[1] = z * x1[0] - x * x1[2]; 
	x2[2] = x * x1[1] - y * x1[0];

	qs = (GAIM_FLOAT)2.0 * s;
	e3la_vec r(v.x + qs * x1[0] +  (GAIM_FLOAT)2.0 * x2[0],
		v.y + qs * x1[1] +  (GAIM_FLOAT)2.0 * x2[1],
		v.z + qs * x1[2] +  (GAIM_FLOAT)2.0 * x2[2]);
	return r;
}

inline e3la_quat e3la_quat::operator!() const {
	e3la_quat r(s, -x, -y, -z);
	return r;
}

inline e3la_quat& e3la_quat::normalize() {
	GAIM_FLOAT l = (GAIM_FLOAT)sqrt(s * s + x * x + y * y + z * z);
	s /= l; x /= l; y /= l; z /= l; 
	return *this;	
}

inline void e3la_quat::print(const char *text, const char *fmt /* = NULL */) const {
	char buf[1024];
	if (text) printf("%s", text);
	if (fmt == NULL) fmt = "%2.2f";
	sprintf(buf, "s: %s x: %s y: %s z: %s\n", fmt, fmt, fmt, fmt);
	printf(buf, s, x, y, z);
}

inline void e3la_mat::set(char axis, GAIM_FLOAT angle) {
	e3la_quat q;
	q.set(axis, angle);
	set(q);
}

inline void e3la_mat::set(const e3la_quat &q) {
	m11 = (GAIM_FLOAT)2.0 * (q.s * q.s + q.x * q.x) - (GAIM_FLOAT)1.0;
	m12 = (GAIM_FLOAT)2.0 * (q.x * q.y - q.s * q.z);
	m13 = (GAIM_FLOAT)2.0 * (q.x * q.z + q.s * q.y);

	m21 = (GAIM_FLOAT)2.0 * (q.x * q.y + q.s * q.z);
	m22 = (GAIM_FLOAT)2.0 * (q.s * q.s + q.y * q.y) - (GAIM_FLOAT)1.0;
	m23 = (GAIM_FLOAT)2.0 * (q.y * q.z - q.s * q.x);

	m31 = (GAIM_FLOAT)2.0 * (q.x * q.z - q.s * q.y);
	m32 = (GAIM_FLOAT)2.0 * (q.y * q.z + q.s * q.x);
	m33 = (GAIM_FLOAT)2.0 * (q.s * q.s + q.z * q.z) - (GAIM_FLOAT)1.0;
}

inline void e3la_mat::set(const e3la_axis &a) {
	GAIM_FLOAT sinA, cosA, versA;
	sinA = (GAIM_FLOAT)sin(a.a);
	cosA = (GAIM_FLOAT)cos(a.a);
	versA = (GAIM_FLOAT)1.0 - cosA;

	m11 = a.x * a.x * versA + cosA;
	m22 = a.y * a.y * versA + cosA;
	m33 = a.z * a.z * versA + cosA;

	m12 = a.x * a.y * versA - a.z * sinA;
	m23 = a.y * a.z * versA - a.x * sinA;
	m31 = a.x * a.z * versA - a.y * sinA;
	
	m13 = a.x * a.z * versA + a.y * sinA;
	m21 = a.x * a.y * versA + a.z * sinA;
	m32 = a.y * a.z * versA + a.x * sinA;
}

inline void e3la_mat::set(const e3la_euler &ea) {
	e3la_quat q;
	q.set(ea);
	set(q);
}

inline void e3la_mat::identity() {
	m11 = m22 = m33 = 1.0;
	m12 = m13 = m21 = m23 = m31 = m32 = 0.0;
}

inline e3la_mat e3la_mat::operator*(const e3la_mat &m) const {
	e3la_mat r(m11 * m.m11 + m12 * m.m21 + m13 * m.m31,
		m11 * m.m12 + m12 * m.m22 + m13 * m.m32,
		m11 * m.m13 + m12 * m.m23 + m13 * m.m33,
		m21 * m.m11 + m22 * m.m21 + m23 * m.m31,
		m21 * m.m12 + m22 * m.m22 + m23 * m.m32,
		m21 * m.m13 + m22 * m.m23 + m23 * m.m33,
		m31 * m.m11 + m32 * m.m21 + m33 * m.m31,
		m31 * m.m12 + m32 * m.m22 + m33 * m.m32,
		m31 * m.m13 + m32 * m.m23 + m33 * m.m33);
	return r;
}

inline e3la_vec e3la_mat::operator*(const e3la_vec &v) const {
	return e3la_vec(m11 * v.x + m12 * v.y + m13 * v.z,
		m21 * v.x + m22 * v.y + m23 * v.z,
		m31 * v.x + m32 * v.y + m33 * v.z);
}

inline e3la_vec transposeMatrixVectorProduct(const e3la_mat &m, const e3la_vec &v) {
	return e3la_vec(
		m.m11 * v.x + m.m21 * v.y + m.m31 * v.z,
		m.m12 * v.x + m.m22 * v.y + m.m32 * v.z,
		m.m13 * v.x + m.m23 * v.y + m.m33 * v.z);
}

inline e3la_mat e3la_mat::operator!() const {
	e3la_mat r;

	r.m11 = m11; r.m12 = m21; r.m13 = m31;
	r.m21 = m12; r.m22 = m22; r.m23 = m32;
	r.m31 = m13; r.m32 = m23; r.m33 = m33;

	return r;
}

inline void e3la_mat::print(const char *text, const char *fmt /*= NULL*/) const {
	char buf[1024];
	if (text) printf("%s\n", text);
	if (fmt == NULL) fmt = "%2.2f";
	sprintf(buf, "%s %s %s\n%s %s %s\n%s %s %s\n", fmt, fmt, fmt, fmt, fmt, fmt, fmt, fmt, fmt);
	printf(buf, m11, m12, m13, m21, m22, m23, m31, m32, m33);
}


inline void e3la_axis::set(const e3la_mat &m) {
	e3la_quat q;
	q.set(m);
	set(q);
}

inline void e3la_axis::set(const e3la_quat &q) {
	GAIM_FLOAT l = (GAIM_FLOAT)sqrt(q.x * q.x + q.y * q.y + q.z * q.z);
	if (l == 0.0) {
		identity();
	}
	else {
		x = q.x / l;
		y = q.y / l;
		z = q.z / l;
		a = (GAIM_FLOAT)2.0 * (GAIM_FLOAT)atan2(l, q.s);
	}
}

inline void e3la_axis::identity() {
	a = 0.0; 
	x = 1.0; 
	y = z = 0.0;
}

inline e3la_vec e3la_axis::operator*(const e3la_vec &v) const {
	e3la_quat q;
	q.set(*this);
	return q * v;
}

inline void e3la_axis::print(const char *text, const char *fmt /*= NULL*/) const {
	char buf[1024];
	if (text) printf("%s", text);
	if (fmt == NULL) fmt = "%2.2f";
	sprintf(buf, "angle: %s axis x: %s y: %s z: %s\n", fmt, fmt, fmt, fmt);
	printf(buf, a, x, y, z);
}


void s_quat2eulerXYZ(const GAIM_FLOAT q[4], GAIM_FLOAT xyz[3], const GAIM_FLOAT prev[3]);
void s_quat2eulerYZX(const GAIM_FLOAT q[4], GAIM_FLOAT yzx[3], const GAIM_FLOAT prev[3]);
void s_quat2eulerYXZ(const GAIM_FLOAT q[4], GAIM_FLOAT yxz[3], const GAIM_FLOAT prev[3]);

inline void e3la_euler::set(eulerSeqOrder order, const e3la_mat &m, const e3la_euler *pea) {
	e3la_quat q;
	q.set(m);
	set(order, q, pea);
}

inline void e3la_euler::set(eulerSeqOrder order, const e3la_quat &aq, const e3la_euler *pea) {
	GAIM_FLOAT ea[3] = {ea1, ea2, ea3}, q[4] = {aq.s, aq.x, aq.y, aq.z}, qInv[4]; 
	GAIM_FLOAT *prevea = NULL, peac[3], preveaInv[3], eaInv[3];
	if (pea) {
		peac[0] = pea->ea1; peac[1] = pea->ea2; peac[2] = pea->ea3;
		prevea = peac;
	}

	o = order;

	switch (order) {
	case EULER_XYZ:
		s_quat2eulerXYZ(q, ea, prevea);
		return;
	case  EULER_YZX:
		s_quat2eulerYZX(q, ea, prevea);
		return;
	case  EULER_YXZ:
		s_quat2eulerYXZ(q, ea, prevea);
		return;
	default:
		/* invert quaternion */
		qInv[0] = q[0];
		qInv[1] = -q[1];
		qInv[2] = -q[2];
		qInv[3] = -q[3];
		/* invert previous euler angles */
		if (prevea) {
			preveaInv[0] = -prevea[2];
			preveaInv[1] = -prevea[1];
			preveaInv[2] = -prevea[0];
			prevea = preveaInv;
		}
		
		switch (order) {
		case  EULER_ZYX:
			s_quat2eulerXYZ(qInv, eaInv, prevea);
			break;
		case  EULER_XZY:
			s_quat2eulerYZX(qInv, eaInv, prevea);
			break;
		case  EULER_ZXY:
			s_quat2eulerYXZ(qInv, eaInv, prevea);
			break;
		default:
			identity();
			return;
		}

		/* invert euler angles */
		ea[0] = -eaInv[2];
		ea[1] = -eaInv[1];
		ea[2] = -eaInv[0];
	}

	ea1 = ea[0];
	ea2 = ea[1];
	ea3 = ea[2];
}

inline void e3la_euler::identity() {
	o = EULER_XYZ;
	ea1 = ea2 = ea3 = 0.0;
}

inline e3la_vec e3la_euler::operator*(const e3la_vec &v) const {
	e3la_quat q;
	q.set(*this);
	return q * v;
}

#endif /* _E3LA_H_ */
