
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _MESH_H_
#define _MESH_H_

#include <vector>

#include "classes.h"
#include "vertex.h"
#include "face.h"

namespace { // anonymous namspace

// simple helper struct used for bsp tree construction
struct bsp_sort_data {
	int idx;
	p3ga::mv::Float signed_distance;
} bsp_sort_data;

}

class mesh {
public:
	mesh(); ///< null constructor
	mesh(const mesh &m); ///< copy constructor
	~mesh(); ///< destructor

	mesh &operator=(const mesh &m); //< copy operator	

	/// returns m_filename
	inline const std::string &get_filename() const {return m_filename;}

	/// simply sets m_filename (does not load or anything)
	void set_filename(const std::string &s);

	/// returns the number of vertices 
	inline const int get_nb_vertices() const {return (int)m_vertex.size();}
	/// sets the number of vertices 
	inline void set_nb_vertices(int nb) {m_vertex.resize(nb);}
	/// returns the vertex at index \param idx (array bounds not checked)
	inline const vertex &get_vertex(int idx) const {return m_vertex[idx];}
	/// sets the vertex at index \param idx (array bounds checked)
	void set_vertex(int idx, const vertex &v);

	/// sets the number of vertices 
	inline void set_nb_faces(int nb) {m_face.resize(nb);}
	/// returns the number of faces
	inline const int get_nb_faces() const {return (int)m_face.size();}
	/// returns the face at index \param idx (array bounds not checked)
	inline const face &get_face(int idx) const {return m_face[idx];}
	/// sets the face at index \param idx (array bounds not checked)
	void set_face(int idx, const face &f);

	/// returns m_bsp_node
	inline const bsp_node *get_bsp_node() const {return m_bsp_node;}
		
	/// Initializes the mesh.
	void init();

	/// returns m_bounding_sphere
	inline const p3ga::normalizedPoint& get_bounding_sphere_center() const {return m_bounding_sphere_center;}

	inline p3ga::mv::Float get_bounding_sphere_radius() const {return m_bounding_sphere_radius;}

	/**
	Finds the intersection of ray \param R and this mesh.
	Depending on the value of \param find_closest , either a search
	for the closes intersection will be done, or for _any_ intersection.
	For shadow rays, for instance, we only need to know that they intersect
	a model, now what model or where.
	When an intersection is found, and \param find_closest is true, 
	information about that point is stored in \param spt .

	Calls get_mesh()->find_intersection....() to do the actual intersection test.
	*/
	bool find_intersection(const ray &R, surface_point &spt, bool find_closest) const;

private:
	std::string m_filename; ///< the .rtm file this mesh was read from

	/// The vertices of this mesh (the indices of the faces in m_face point into this vector)
	std::vector<vertex> m_vertex;

	/// The faces of this mesh:
	std::vector<face> m_face;

	/// Pointer to the BSP tree for this mesh (destruction of object managed by this class)
	bsp_node *m_bsp_node;

	/// Bounding sphere (center) for all vertices
	p3ga::normalizedPoint m_bounding_sphere_center;
	/// Bounding sphere (radius) for all vertices
	p3ga::mv::Float m_bounding_sphere_radius;

	/// Initializes all faces; called when a mesh has just been read from file
	void init_faces();
	/// Compute the bounding sphere; called when a mesh has just been read from file
	void compute_bounding_sphere();
	/// Constructs the BSP tree for this mesh; called when a mesh has just been read from file
	void construct_bsp_tree();
	/// Computes a local 'basis' for the texture for each face; called when a mesh has just been read from file
	void compute_local_texture_basis();


	/// Helper function for BSP tree construction 
	bsp_node *construct_bsp_tree(int base_plane_idx, unsigned int nb_vertices, struct bsp_sort_data*vertex, unsigned int desired_vertices_per_partition, std::vector<unsigned int> lastSizes) const;

	/// Another helper function for BSP tree construction 
	bsp_node *construct_bsp_leaf(unsigned int nb_vertices, struct bsp_sort_data*vertex) const;

	/**
	Same as find_intersection_brute_force(), but uses BSP tree to speed up test.
	The parameter \param pp contains the initial point pair that we have to test against
	the BSP tree.
	*/
	bool find_intersection_bsp(const p3ga::normalizedPoint &start_pt, const p3ga::normalizedPoint &end_pt, const ray &R, surface_point &spt, bool find_closest) const;

};

inline const vertex &face::get_vtx(int idx, const mesh &m) const {return m.get_vertex(get_vtx_idx(idx));}



#endif /* _MESH_H_ */
