
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <math.h>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <GL/gl.h>

#include "c3ga_util.h"
#include "light_source.h"
#include "ray.h"
#include "scene.h"
#include "surface_point.h"
#include "../shared.src/string_util.h"
#include "../shared.src/intersection_info.h"


light_source::light_source() :
	m_spot_exponent(0.0f), 
	m_spot_cutoff(100.0f), 
	m_constant_attenuation(1.0f), 
	m_linear_attenuation(0.0f), 
	m_quadratic_attenuation(0.0f)  {

}

light_source::light_source(const p3la_tm &tm, const p3la_tm &itm,
		const color &ambient_color,
		const color &diffuse_color,
		const color &specular_color,
		GAIM_FLOAT spot_exponent,
		GAIM_FLOAT spot_cutoff,
		GAIM_FLOAT constant_attenuation,
		GAIM_FLOAT linear_attenuation,
		GAIM_FLOAT quadratic_attenuation) : locator(tm, itm), 
	m_ambient_color(ambient_color),
	m_diffuse_color(diffuse_color),
	m_specular_color(specular_color),
	m_spot_exponent(spot_exponent), 
	m_spot_cutoff(spot_cutoff), 
	m_constant_attenuation(constant_attenuation), 
	m_linear_attenuation(linear_attenuation), 
	m_quadratic_attenuation(quadratic_attenuation) {
}

light_source::light_source(const light_source &l) : locator(l),
	m_ambient_color(l.m_ambient_color),
	m_diffuse_color(l.m_diffuse_color),
	m_specular_color(l.m_specular_color),
	m_spot_exponent(l.m_spot_exponent), 
	m_spot_cutoff(l.m_spot_cutoff), 
	m_constant_attenuation(l.m_constant_attenuation), 
	m_linear_attenuation(l.m_linear_attenuation), 
	m_quadratic_attenuation(l.m_quadratic_attenuation) {
}

light_source::~light_source() {
}

light_source &light_source::operator=(const light_source &l) {
	if (this != & l) {
		locator::operator=(l);
		m_ambient_color = l.m_ambient_color;
		m_diffuse_color = l.m_diffuse_color;
		m_specular_color = l.m_specular_color;
		m_spot_exponent = l.m_spot_exponent;
		m_spot_cutoff = l.m_spot_cutoff;
		m_constant_attenuation = l.m_constant_attenuation;
		m_linear_attenuation = l.m_linear_attenuation;
		m_quadratic_attenuation = l.m_quadratic_attenuation;
	}
	return *this;
}


bool light_source::can_scale() const {
	return false;
}


void light_source::set_ambient_color(const color &c) {
	m_ambient_color = c;
}
	
void light_source::set_diffuse_color(const color &c) {
	m_diffuse_color = c;
}

void light_source::set_specular_color(const color &c) {
	m_specular_color = c;
}

void light_source::set_color(const std::string colorname, const color &c) {
	if ((colorname[0] == 'a') || (colorname[0] == 'A'))
		set_ambient_color(c);
	else if ((colorname[0] == 'd') || (colorname[0] == 'D'))
		set_diffuse_color(c);
	else if ((colorname[0] == 's') || (colorname[0] == 'S'))
		set_specular_color(c);
}

void  light_source::set_spot_exponent(GAIM_FLOAT f) {
	m_spot_exponent = f;
}

void light_source::set_spot_cutoff(GAIM_FLOAT c) {
	if (c < 0.0f) m_spot_cutoff = 0.0f;
	else if (c > 100.0f) m_spot_cutoff = 100.0f;
	else m_spot_cutoff = c;
}

void  light_source::set_constant_attenuation(GAIM_FLOAT f) {
	m_constant_attenuation = f;
}

void  light_source::set_linear_attenuation(GAIM_FLOAT f) {
	m_linear_attenuation = f;
}

void  light_source::set_quadratic_attenuation(GAIM_FLOAT f) {
	m_quadratic_attenuation = f;
}

color light_source::shade(const ray &R, const scene &S, const intersection_info *II, const surface_point &spt) const {
	// perform basic 'OpenGL-style' lighting computations (+ a shadow check)...
	// get 'viewing direction' towards the light as a unit free vector
	const p3la_vec view_dir(-R.get_direction());

	// get direction from surface point to light as unit free vector
	p3la_vec light_dir = (get_translation() - spt.get_pt()).normalize();

	// evaluate the contribution of every light and add it to 'c'
	// is the light visible?
	bool light_visble;
	if (II) light_visble = (II->get_light_visible(S.get_light_index(this)));
	else {
		surface_point tmp_spt;
		// spawn a shadow ray:
		light_visble = !S.find_intersection(ray(spt.get_pt(), light_dir), tmp_spt, false); // false means: don't search for closest point
	}

//	light_visble = true;

	color C; // this is where the shading computations are accumulated

	if (light_visble) { // perform diffuse & specular lighting computations
		// compute the diffuse term
//		printf("Att:  %s\n", spt.get_att().string("%f"));
		float fd = (float)(spt.get_pl().normal() % light_dir);
		if (fd < 0.0f) fd = 0.0f;
		const float terminator_hack_threshold = 0.1f; // hack to limit the 'terminator effect'
		if (fd < terminator_hack_threshold) fd = (fd * fd * fd) / (terminator_hack_threshold * terminator_hack_threshold);
		//printf("fd = %f,\n", fd);

		// add diffuse term to 'C'
		C += fd * get_diffuse_color() * spt.get_diffuse_color();

		// compute the specular term
		p3la_vec half_dir = (view_dir + light_dir).normalize();
//		printf("view_dir = %s,\n", view_dir.string());
//		printf("light_dir = %s,\n", light_dir.string());
		float fs = (float)(spt.get_pl().normal() % half_dir);
//		printf("%f %f %f\n", fs, spt.get_shininess(), (float)pow(fs, (float)spt.get_shininess()));
		fs = (fs < 0.0f) ? 0.0f : (float)pow(fs, (float)spt.get_shininess());

		// add specular term to 'C'
		C += fs * get_specular_color() * spt.get_specular_color();

//		printf("l spec: %s\n", get_specular_color().to_string().c_str());
//		printf("spt spec: %s\n", spt.get_specular_color().to_string().c_str());

	}

	// add ambient term to 'C'
	C += get_ambient_color() * spt.get_ambient_color();

	// spotlight (matches OpenGL: it applies spotlight to ambient too).
	if (get_spot_cutoff() < 95.0) {
		// get cutoff in radians; get cosine of cutoff
		float cutoff = (get_spot_cutoff() < 90.0f) ? (float)get_spot_cutoff() : 90.0f;
		float cutoff_rad = (float)M_PI * cutoff / 180.0f;

		// compute the direction the spot is shining in, as a free vector:
		p3la_vec spot_dir = transform_direction(p3la_vec(0.0f, 0.0f, 1.0f));

		// compute the cosine of the angle between the light_dir and spot_dir
		float cos_angle = (float)(light_dir % spot_dir);
		// the point lies outside the cone
		if (cos_angle < cos(cutoff_rad)) cos_angle = 0.0f; 
		// apply spotlight to color
		C *= (float)pow(cos_angle, (float)get_spot_exponent()); 
	}

	
//	Attenuation:
//	-compute distance of light source to surface point
//	-scale 'C' by the attenuation factor
	float distance = (float)(get_translation() - spt.get_pt()).length();
	C *= 1.0f / ((float)get_constant_attenuation() +
		(float)get_linear_attenuation() * distance +
		(float)get_quadratic_attenuation() * distance * distance);

	return C;
}


