
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _LIGHT_SOURCE_H_
#define _LIGHT_SOURCE_H_

#include "classes.h"
#include "locator.h"
#include "../shared.src/color.h"

class light_source : public locator {
public:
	light_source(); //< null constructor
	light_source(const p3la_tm &tm, const p3la_tm &itm,
		const color &ambient_color,
		const color &diffuse_color,
		const color &specular_color,
		GAIM_FLOAT spot_exponent,
		GAIM_FLOAT spot_cutoff,
		GAIM_FLOAT constant_attenuation,
		GAIM_FLOAT linear_attenuation,
		GAIM_FLOAT quadratic_attenuation); //< constructor; sets all member variables
	light_source(const light_source &l); //< copy constructor
	~light_source(); //< destructor

	light_source &operator=(const light_source &l); //< assignment operator

	/// returns m_spot_exponent
	inline GAIM_FLOAT get_spot_exponent() const {return m_spot_exponent;}
	/// returns m_spot_cutoff (degrees)
	inline GAIM_FLOAT get_spot_cutoff() const {return m_spot_cutoff;}
	/// returns m_spot_cutoff (radians)
	inline GAIM_FLOAT get_spot_cutoff_rad() const {return M_PI * m_spot_cutoff / 180.0;}
	/// returns m_constant_attenuation
	inline GAIM_FLOAT get_constant_attenuation() const {return m_constant_attenuation;}
	/// returns m_linear_attenuation
	inline GAIM_FLOAT get_linear_attenuation() const {return m_linear_attenuation;}
	/// returns m_quadratic_attenuation
	inline GAIM_FLOAT get_quadratic_attenuation() const {return m_quadratic_attenuation;}

	/// sets m_spot_exponent
	void set_spot_exponent(GAIM_FLOAT f);
	/// sets m_spot_cutoff
	void set_spot_cutoff(GAIM_FLOAT f);
	/// sets m_constant_attenuation
	void set_constant_attenuation(GAIM_FLOAT f);
	/// sets m_linear_attenuation
	void set_linear_attenuation(GAIM_FLOAT f);
	/// sets m_quadratic_attenuation
	void set_quadratic_attenuation(GAIM_FLOAT f);


	/// returns m_ambient_color
	const color &get_ambient_color() const {return m_ambient_color;}
	/// returns m_diffuse_color
	const color &get_diffuse_color() const {return m_diffuse_color;}
	/// returns m_specular_color
	const color &get_specular_color() const {return m_specular_color;}

	/// sets m_ambient_color
	void set_ambient_color(const color &c);
	/// sets m_diffuse_color
	void set_diffuse_color(const color &c);
	/// sets m_specular_color
	void set_specular_color(const color &c);

	/// Sets color of 'ambient', 'diffuse' or 'specular'
	void set_color(const std::string colorname, const color &c);


	/// activates this light in OpenGL. \param OpenGL_light_idx specifies what OpenGL light to use.
	void to_OpenGL(int OpenGL_light_idx) const;

	/**
	Draws this light as a box with 'flaps'.
	The flaps indicate the size of the spotlight
	*/
	void to_OpenGL_model() const;

	/// a light source can not scale, so this function returns false
	virtual bool can_scale() const;

	/**
	Shades the surface point \param spt as seen from ray \param R .
	\param S is used to determine whether the light is visible from the surface
	point. If \param II is not NULL, it used as a 'shortcut' to determine 
	if the light is visible.
	*/
	color shade(const ray &R, const scene  &S, const intersection_info *II, const surface_point &spt) const;


private:
	// allow for 'spotlight' toggle in UI
	GAIM_FLOAT m_spot_exponent; ///< spot exponent in OpenGL shading model
	GAIM_FLOAT m_spot_cutoff; ///< 'width' of spot in OpenGL shading model (in degrees [0, 90])
	GAIM_FLOAT m_constant_attenuation;
	GAIM_FLOAT m_linear_attenuation;
	GAIM_FLOAT m_quadratic_attenuation;

	color m_ambient_color; ///< ambient color of this light
	color m_diffuse_color; ///< diffuse color of this light
	color m_specular_color; ///< specular color of this light

};

#endif /* _LIGHT_SOURCE_H_ */
