
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _COLOR_H_
#define _COLOR_H_

#include <string>

// include gl.h
/*#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#include <GL/gl.h>*/

class color {
public:
	/// null constructor
	inline color() {
		m_rgb[0] = m_rgb[1] = m_rgb[2] = 0.0;
		m_rgb[3] = 1.0;
	}

	/// constructor that sets all values
	inline color(float r, float g, float b) {
		set(r, g, b);
	}

	/// constructor that sets all values
	inline color(float r, float g, float b, float a) {
		set(r, g, b, a);
	}

	/// constructor that sets all values from an array
	inline color(const float rgb[4]) {
		set4v(rgb);
	}

	/// constructor that sets all values from an array of unsigned ints
	inline color(const unsigned char rgb[3]) {
		set((float)rgb[0] / 255.0f, (float)rgb[1] / 255.0f, (float)rgb[2] / 255.0f, 1.0f);
	}

	/// copy constructor
	inline color(const color &c) {
		set(c);
	}

	/// destructor
	~color() {}

	/// assignment operator
	color &operator=(const color &c) {
		if (this != &c) 
			set(c);
		return *this;
	}

	/// sets the values of this color
	inline void set(float r, float g, float b) {
		m_rgb[0] = r;
		m_rgb[1] = g; 
		m_rgb[2] = b;
		m_rgb[3] = 1.0;
	}

	/// sets the values of this color
	inline void set(float r, float g, float b, float a) {
		m_rgb[0] = r;
		m_rgb[1] = g; 
		m_rgb[2] = b;
		m_rgb[3] = a;
	}

	/// sets the values of this color from an array
	inline void set3v(const float rgb[3]) {
		m_rgb[0] = rgb[0];
		m_rgb[1] = rgb[1]; 
		m_rgb[2] = rgb[2];
		m_rgb[3] = 1.0f;
	}

	/// sets the values of this color from an array
	inline void set4v(const float rgb[4]) {
		m_rgb[0] = rgb[0];
		m_rgb[1] = rgb[1]; 
		m_rgb[2] = rgb[2];
		m_rgb[3] = rgb[3];
	}

	/// sets the values of this color from another color
	inline void set(const color &c) {
		m_rgb[0] = c.m_rgb[0];
		m_rgb[1] = c.m_rgb[1]; 
		m_rgb[2] = c.m_rgb[2];
		m_rgb[3] = c.m_rgb[3];
	}

	/// adds \param c to this color
	inline void add(const color &c) {
		m_rgb[0] += c.m_rgb[0]; 
		m_rgb[1] += c.m_rgb[1]; 
		m_rgb[2] += c.m_rgb[2];
		m_rgb[3] += c.m_rgb[3];
	}

	/// returns the grayscale value of this color
	inline float grayscale() {
		return (m_rgb[0] + m_rgb[1] + m_rgb[2]) / 3.0f;
	}

	/**
	clamps the 3 color values between \param min_val and \param max_val .
	Returns true when the value was actually clamped
	*/
	bool clamp(float min_val, float max_val) {
		bool clamped = false;
		for (int i = 0; i < 3; i++) {
			if (m_rgb[i] < min_val) {
				m_rgb[i] = min_val;
				clamped = true;
			}
			else if (m_rgb[i] > max_val) {
				m_rgb[i] = max_val;
				clamped = true;
			}
		}
		return clamped;
	}

	/// Scales this color by \param factor
	inline void scale(float factor) {
		m_rgb[0] *= factor; 
		m_rgb[1] *= factor; 
		m_rgb[2] *= factor;
//		m_rgb[3] *= factor;
	}

	/// adds \param c to this
	inline color &operator+=(const color &c) {
		m_rgb[0] += c.m_rgb[0];
		m_rgb[1] += c.m_rgb[1];
		m_rgb[2] += c.m_rgb[2];
		m_rgb[3] += c.m_rgb[3];
		return *this;
	}

	/// Scales this color by \param factor
	inline color &operator*=(float factor) {
		scale(factor);
		return *this;
	}

	/// divides this by \param f
	inline color &operator/=(float f) {
		m_rgb[0] /= f;
		m_rgb[1] /= f;
		m_rgb[2] /= f;
		m_rgb[3] /= f;
		return *this;
	}

	/// calls 'glColor3f(m_rgb)'
	void glColor() const; 
	/// calls 'glMaterialfv(\param side, \param parm, m_rgb)'
	void glMaterial(int side, int parm) const; 
	/// calls 'glLightfv(\param idx, \param parm, m_rgb)'
	void glLight(int idx, int parm) const; 
	/// calls 'glLightModelfv(\param parm, m_rgb)'
	void glLightModel(int parm) const; 

	/// returns the red component of this color
	inline float r() const {return m_rgb[0];}
	/// returns the green component of this color
	inline float g() const {return m_rgb[1];}
	/// returns the blue component of this color
	inline float b() const {return m_rgb[2];}
	/// returns the alpha component of this color
	inline float a() const {return m_rgb[3];}

	/// Turns this object into a string (e.g., for storing a scene to file)
	std::string to_string() const;

private:
	float m_rgb[4]; ///< color red, green, blue, alpha values
};

/// returns the component-wise addition of \param c1 and \param c2 .
inline color operator+(const color &c1, const color &c2) {
	return color(c1.r() + c2.r(), c1.g() + c2.g(), c1.b() + c2.b(), c1.a() + c2.a());
}

/// returns the component-wise multiplication of \param c1 and \param c2 .
inline color operator*(const color &c1, const color &c2) {
	return color(c1.r() * c2.r(), c1.g() * c2.g(), c1.b() * c2.b(), c1.a() * c2.a());
}

/// returns the scalar multiplication of \param s and \param c .
inline color operator*(float s, const color &c) {
	return color(s * c.r(), s * c.g(), s * c.b(), s * c.a());
}

#endif /* _COLOR_H_ */

